# %%
# IMPORT EXTERNAL LIBRARIES AND PACKAGES

import os
from collections import Counter
from itertools import product

import geopandas as gpd
import numpy as np
import pandas as pd
import seaborn as sns
from hdbscan import HDBSCAN
from sklearn.cluster import KMeans
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import (
    calinski_harabasz_score,
    davies_bouldin_score,
    silhouette_score,
)
from sklearn.preprocessing import RobustScaler
from sklearn import decomposition as dec
from hopkins import hopkins

from tqdm import tqdm

from joblib import Parallel, delayed

import paths

# %%
# Define columns and other constant values

quantile_range = (20, 80)

# air quality
aq_raw_cols = ["Area_r1", "Area_r2", "Area_r3", "Area_r4", "Area_r6", "Area_r9"]
aq_nor_cols = [
    "P_Area_r1",
    "P_Area_r2",
    "P_Area_r3",
    "P_Area_r4",
    "P_Area_r6",
    "P_Area_r9",
]

# carbon
# absorption
co2_raw_cols = ["CO2"]
co2_nor_cols = ["kgCO2/m²"]

# emmissions
ems = ["gasolio_em", "gpl_em", "ee_em", "metano_em", "legno_em", "tot_em", "c_ee_em"]
co2_em_raw_cols = ["finale"]
co2_em_nor_cols = ["finale/m²"]

# flora fauna and habitat
ffh_raw_cols = ["F_Area_m2"]
ffh_nor_cols = ["P_Prot_are"]

# spatial
spt_ag_raw_cols = ["Area_300m_", "Area_750m_"]
spt_ag_nor_cols = ["P_area_300", "P_area_750"]

spt_wk_nor_cols = ["MEAN_Resul"]

# temporary
tem_raw_cols = ["Area_diff"]
tem_nor_cols = ["P_Area_dif"]

# thermal
thm_nor_cols = ["mean_norm"]

# socio-demographic
socio_abs_cols = [
    "Maschi",
    "Femmine",
    "Totale popolazione",
    "0-14",
    "15-64",
    "65 e oltre",
    "indice di vecchiaia",
    "0-17",
    "Stranieri",
    "Media componenti",
    "Totale famiglie con figli",
    "Totale famiglie senza figli",
    "Totale famiglie",
    "Single",
]
socio_perc_cols = [
    "%65+",
    "%minori",
    "%stranieri",
    "%con figli",
    "%senza figli",
    "%Single",
]

sez_cols = ["geometry", "SEZ", "SEZ2011", "COD_REG", "COD_ISTAT", "PRO_COM"]

# %%
## SELECT COLUMNS TO BE USED

cols = [
    # air-quality: % risk based on SVF and road classification
    "P_Area_r1",
    "P_Area_r2",
    "P_Area_r3",
    "P_Area_r4",
    "P_Area_r6",
    "P_Area_r9",
    # carbon: absorption and emmissions
    "kgCO2/m²",
    "finale/m²",
    # FFH: % protected area, % to 5 min to prot. areas
    "P_Prot_are",
    # spatial: % of area within a distance from protected areas and walkability index
    "P_area_300",
    # "P_area_750",
    "MEAN_Resul",
    # temporal: % area changed
    "P_Area_dif",
    # thermal
    "mean_norm",
    # socio demographic
    # "%65+",
    # "%minori",
    "indice di vecchiaia",
    "%stranieri",
    "%con figli",
    "%senza figli",
    "%Single",
]

# define weights
wcols = np.array(
    [
        # air-quality: % risk based on SVF and road classification
        1.0 / 6.0,
        1.0 / 6.0,
        1.0 / 6.0,
        1.0 / 6.0,
        1.0 / 6.0,
        1.0 / 6.0,
        # carbon: absorption and emmissions
        1.0 / 2.0,
        1.0 / 2.0,
        # FFH: % protected area, % to 5 min to prot. areas
        1.0,
        # spatial: % of area within a distance from protected areas and walkability index
        1.0 / 2.0,
        # "P_area_750",
        1.0 / 2.0,
        # temporal: % area changed
        1.0,
        # thermal
        1.0,
        # socio demographic
        # "%65+",
        # "%minori",
        1.0 / 5.0,
        1.0 / 5.0,
        1.0 / 5.0,
        1.0 / 5.0,
        1.0 / 5.0,
    ],
    dtype=float,
)


# %%
# Define function to read files


def read_data():
    aq_scr = gpd.read_file(paths.aq_scr_shp)
    co2_ab = gpd.read_file(paths.co2_ab_shp)
    co2_ab["kgCO2/m²"] = co2_ab["CO2"] / co2_ab.area
    co2_em = gpd.read_file(paths.co2_em_shp)
    co2_em["finale/m²"] = co2_em["finale"] / co2_em.area
    ffh_pa = gpd.read_file(paths.ffh_pa_shp)
    spt_ag = gpd.read_file(paths.spt_ag_shp)
    spt_wk = gpd.read_file(paths.spt_wk_shp)
    tem_ss = gpd.read_file(paths.tem_ss_shp)
    thm_sh = gpd.read_file(paths.thm_sh_shp)

    socio_pk = gpd.read_file(paths.socio_gpkg, driver="GPKG")
    # fix column
    res = []
    for val in socio_pk["indice di vecchiaia"]:
        try:
            v = float(val.strip().replace(",", "."))
        except Exception:
            v = np.nan
        res.append(v)
    socio_pk["indice di vecchiaia"] = np.array(res, dtype=float)

    raw = pd.concat(
        [
            co2_ab.loc[:, sez_cols],
            aq_scr.loc[:, aq_raw_cols + aq_nor_cols],
            co2_ab.loc[:, co2_raw_cols + co2_nor_cols],
            co2_em.loc[:, co2_em_raw_cols + co2_em_nor_cols + ems],
            ffh_pa.loc[:, ffh_raw_cols + ffh_nor_cols],
            spt_ag.loc[:, spt_ag_raw_cols + spt_ag_nor_cols],
            spt_wk.loc[:, spt_wk_nor_cols],
            tem_ss.loc[:, tem_raw_cols + tem_nor_cols],
            thm_sh.loc[:, thm_nor_cols],
            socio_pk.loc[:, socio_abs_cols + socio_perc_cols],
        ],
        axis="columns",
    )
    return raw


def raw2scaled(
    raw, quantile_range=quantile_range, start="Area_r1", end=None, scaler=None
):
    if scaler is None:
        scaler = RobustScaler(
            with_centering=True, with_scaling=True, quantile_range=quantile_range
        )
    scaled = raw.loc[:, sez_cols].copy()
    xr = raw.loc[:, start:end]
    xs = scaler.fit_transform(xr)
    for col, xarr in zip(xr.columns, xs.T):
        scaled[col] = xarr
    return scaled


# %%
# define function to compute Hopkins statistics

"""
def hopkins_statistic(X):
    '''Function to compute hopkins's statistic for the dataframe X'''
    # X = X.values  # convert dataframe to a numpy array
    sample_size = int(X.shape[0] * 0.10)  # 0.05 (5%) based on paper by Lawson and Jures

    # a uniform random sample in the original data space
    X_uniform_random_sample = uniform(
        X.min(axis=0), X.max(axis=0), (sample_size, X.shape[1])
    )

    # a random sample of size sample_size from the original data X
    random_indices = sample(range(0, X.shape[0], 1), sample_size)
    X_sample = X[random_indices]

    # initialise unsupervised learner for implementing neighbor searches
    neigh = NearestNeighbors(n_neighbors=2)
    nbrs = neigh.fit(X)

    # u_distances = nearest neighbour distances from uniform random sample
    u_distances, u_indices = nbrs.kneighbors(X_uniform_random_sample, n_neighbors=2)
    u_distances = u_distances[:, 0]  # distance to the first (nearest) neighbour

    # w_distances = nearest neighbour distances from a sample of points from original data X
    w_distances, w_indices = nbrs.kneighbors(X_sample, n_neighbors=2)
    # distance to the second nearest neighbour (as the first neighbour will be the point itself, with distance = 0)
    w_distances = w_distances[:, 1]

    u_sum = np.sum(u_distances)
    w_sum = np.sum(w_distances)

    # compute and return hopkins' statistic
    H = u_sum / (u_sum + w_sum)
    return H


def compute_hopkins_stats(X, n=50):
    hs = pd.Series([hopkins_statistic(X) for _ in range(n)], name="hopkins")
    return hs.describe()
"""


def print_hopkins(mname, hopkins):
    print(
        f"{mname} pre-processed data have a Hopkins value of: "
        f"avg: {hopkins['mean']:6.4f} +/- {hopkins['std']:6.4f} | "
        f"min: {hopkins['min']:6.4f} | "
        f"25%: {hopkins['25%']:6.4f} | "
        f"50%: {hopkins['50%']:6.4f} | "
        f"75%: {hopkins['75%']:6.4f} | "
        f"max: {hopkins['max']:6.4f} | "
    )


# %%
# define cluster metrics


def compute_metrics(model, data, labels, alpha_k=0.02):
    k = len(set(labels)) - 1
    # avoid to compute metrics for outliers, rm: -1
    valid = labels >= 0
    labels = labels[valid]
    data = data.loc[valid, :]
    if hasattr(model, "inertia_"):
        inertia_o = np.square((data.values - data.values.mean(axis=0))).sum()
        inertia = model.inertia_
        scaledi = inertia / inertia_o + alpha_k * k
    else:
        inertia_o = None
        inertia = None
        scaledi = None

    if k <= (len(labels) - 1) and k >= 2:
        # Number of labels is 1. Valid values are 2 to n_samples - 1
        sil = silhouette_score(data, labels)
        dvb = davies_bouldin_score(data, labels)
        cal = calinski_harabasz_score(data, labels)
        return k, inertia, scaledi, sil, dvb, cal
    return k, inertia, scaledi, None, None, None


# %%
# Define class to give feature importances
# taken from MIT:
# https://github.com/YousefGh/kmeans-feature-importance


class KMeansInterp(KMeans):
    def __init__(
        self, ordered_feature_names, feature_importance_method="wcss_min", **kwargs
    ):
        super(KMeansInterp, self).__init__(**kwargs)
        self.feature_importance_method = feature_importance_method
        self.ordered_feature_names = ordered_feature_names

    def fit(self, X, y=None, sample_weight=None):
        super().fit(X=X, y=y, sample_weight=sample_weight)

        if not len(self.ordered_feature_names) == self.n_features_in_:
            raise Exception(
                f"Model is fitted on {self.n_features_in_} but ordered_feature_names = {len(self.ordered_feature_names)}"
            )

        if self.feature_importance_method == "wcss_min":
            self.feature_importances_ = self.get_feature_imp_wcss_min()
        elif self.feature_importance_method == "unsup2sup":
            self.feature_importances_ = self.get_feature_imp_unsup2sup(X)
        else:
            raise Exception(
                f" {self.feature_importance_method}"
                + "is not available. Please choose from  ['wcss_min' , 'unsup2sup']"
            )

        return self

    def get_feature_imp_wcss_min(self):
        labels = self.n_clusters
        centroids = self.cluster_centers_
        centroids = np.vectorize(lambda x: np.abs(x))(centroids)
        sorted_centroid_features_idx = centroids.argsort(axis=1)[:, ::-1]

        cluster_feature_weights = {}
        for label, centroid in zip(range(labels), sorted_centroid_features_idx):
            ordered_cluster_feature_weights = centroids[label][
                sorted_centroid_features_idx[label]
            ]
            ordered_cluster_features = [
                self.ordered_feature_names[feature] for feature in centroid
            ]
            cluster_feature_weights[label] = list(
                zip(ordered_cluster_features, ordered_cluster_feature_weights)
            )

        return cluster_feature_weights

    def get_feature_imp_unsup2sup(self, X):
        try:
            from sklearn.ensemble import RandomForestClassifier
        except ImportError as IE:
            print(IE.__class__.__name__ + ": " + IE.message)
            raise Exception(
                "Please install scikit-learn. "
                + "'unsup2sup' method requires using a classifier"
                + "and depends on 'sklearn.ensemble.RandomForestClassifier'"
            )

        cluster_feature_weights = {}
        for label in range(self.n_clusters):
            binary_enc = np.vectorize(lambda x: 1 if x == label else 0)(self.labels_)
            clf = RandomForestClassifier()
            clf.fit(X, binary_enc)

            sorted_feature_weight_idxes = np.argsort(clf.feature_importances_)[::-1]
            ordered_cluster_features = np.take_along_axis(
                np.array(self.ordered_feature_names),
                sorted_feature_weight_idxes,
                axis=0,
            )
            ordered_cluster_feature_weights = np.take_along_axis(
                np.array(clf.feature_importances_), sorted_feature_weight_idxes, axis=0
            )
            cluster_feature_weights[label] = list(
                zip(ordered_cluster_features, ordered_cluster_feature_weights)
            )
        return cluster_feature_weights


def reshapedf(df):
    ndf = sel_raw.unstack().reset_index()
    ndf.columns = ["features", "stats units", "values"]
    return ndf


def get_cluster_colors(clusters, palette="hls"):
    # prepare colors
    clss = [c for c in clusters if c >= 0]
    colors = {
        cl: clr
        for cl, clr in zip(
            clss,
            sns.color_palette(palette, len(clss)),
        )
    }
    colors[-1] = (0.5, 0.5, 0.5)
    clrs = pd.DataFrame(colors, index=["R", "G", "B"]).T
    clrs.index.name = "cluster"
    clrs.sort_index(inplace=True)
    return clrs


def feature_importances(data, labels, label):
    clf = RandomForestClassifier(max_depth=10, n_estimators=500, random_state=1)
    clf.fit(data.values, labels)
    return pd.DataFrame(
        clf.feature_importances_,
        index=pd.Series(data.columns, name="features"),
        columns=[
            label,
        ],
    ).sort_values(label, ascending=False)


def exec_model(model, opts, data):
    # apply the model
    clst = model["cls"](**opts)
    labels = clst.fit_predict(data)

    # apply all the post-processing actions
    for post in model.get("post-processing", []):
        post(clst, data, label)
    return clst, labels


def cls_counter(labels):
    clcount = pd.DataFrame(Counter(labels).most_common(), columns=["cluster", "count"])
    tot = clcount["count"].sum()
    clcount["%"] = clcount["count"].astype(float) * 100.0 / tot
    perc = clcount.loc[clcount["cluster"] >= 0, "%"].sum()
    clcount.sort_values(by="%", ascending=False, inplace=True)
    if round(perc, 2) > 100.0:
        print("#" * 60)
        print(f"WARNING: percentage > 100%: {perc!r}")
        print(clcount)
        print("#" * 60)
    return clcount, perc


def report(data, dname, models, cldf, rdir):
    def worker(mname, model, opts, tdata):
        label = model["label"](opts)
        clst, labels = exec_model(model, opts, tdata)
        score = compute_metrics(clst, tdata, labels, alpha_k=0.02)
        clcount, perc = cls_counter(labels)
        if score[0] > 1:
            n_of_cls1 = (clcount["count"] == 1).sum()
            mean_el_per_cl = clcount.loc[0:, "count"].mean()
            std_el_per_cl = clcount.loc[0:, "count"].std()
        else:
            n_of_cls1 = None
            mean_el_per_cl = None
            std_el_per_cl = None
        long_label = f"k{len(clcount):03d}_perc{perc:05.1f}_{label}"
        xcores = [
            mname,
            label,
            long_label,
            perc,
            n_of_cls1,
            mean_el_per_cl,
            std_el_per_cl,
        ] + [s for s in score]
        if rdir is not None:
            clcount.to_excel(rdir / f"{long_label}.xlsx")
        fimp = feature_importances(tdata, labels, label)
        return long_label, labels, xcores, fimp

    num_opts = sum(len(m["opts"]) for m in models.values())
    print(f"Exploring the cluster space with {len(models)} models and {num_opts} tasks")

    # compute Hopkins per pre-processing chain per model
    # A statistical test which allow to guess if the data
    # follow an uniform distribution. If the test is positve
    # (an hopkins score which tends to 0) it means that the
    # data is not uniformly distributed. Hence clustering
    # can be useful to classify the observations. However,
    # if the score is too high (above 0.3 for exemple);
    # the data is uniformly distributed and clustering can’t
    # be really useful for the problem at hand.
    trans = {}
    for mname, model in models.items():
        tdata = data.copy()
        # apply all the transformations
        for pre in model.get("pre-processing", []):
            tdata = pre.fit_transform(tdata)

        # avoit to pre-process the data for every attempt
        trans[mname] = tdata
        # compute Hopkins statistics to test the clusterability of the daset
        hop = hopkins(tdata, sampling_size=150)
        print(f"{mname}, Hopkins: {hop:.5f}")
        # print_hopkins(mname, hopkins)

    pll = Parallel(n_jobs=-1, verbose=10)
    tasks = [
        (mname, model, opts, trans[mname])
        for mname, model in models.items()
        for opts in model["opts"]
    ]

    # DEBUG: on
    # mname, model, opts, df = tasks[0]
    # worker(mname, model, opts, df)
    # DBEUG: off

    res = pll(
        delayed(worker)(mname, model, opts, trans[mname])
        for mname, model, opts, df in tqdm(tasks)
    )

    scores, fimps = [], []
    for long_label, labels, xcores, fimp in res:
        scores.append(xcores)
        fimps.append(fimp)
        cldf.loc[data.index, long_label] = labels

    sc = pd.DataFrame(
        scores,
        columns=[
            "model",
            "label",
            "long_label",
            "% covered by clusters",
            "n_of_cl1",
            "mean_el_per_cl",
            "std_el_per_cl",
            "n_clusters",
            "inertia",
            "scaled_inertia",
            "silhouette",
            "davies bouldin",
            "calinski harabasz",
        ],
    )
    fi = pd.concat(fimps, axis=1)
    return sc, fi


class ApplyW:
    def __init__(self, weights):
        self.weights = weights

    def fit_transform(self, X, *args, **kwargs):
        return X * self.weights


# %%

if __name__ == "__main__":
    # %%
    # read and transform data
    raw = read_data()
    raw.to_file("raw.geojson", driver="GeoJSON")

    # use selected columns
    selected = raw.loc[:, cols].dropna()

    # scaled = raw2scaled(raw, scaler=None)
    # scaled.to_file("scaled.geojson", driver="GeoJSON")

    # weighted = selected * wcols
    apply_weight = ApplyW(wcols)

    # %%
    # Explore clustering pre-processing

    preprocs = {
        # "Robust&Weight20-80+PCA": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.PCA(),
        # ),
        # "Robust&Weight20-80+PCAmle": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.PCA(n_components="mle", svd_solver="full"),
        # ),
        # "Robust&Weight20-80+PCArandomize": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.PCA(svd_solver="randomized", whiten=True),
        # ),
        # "Robust&Weight20-80+ICAwarn": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.FastICA(whiten="warn"),
        # ),
        # "Robust&Weight20-80+ICAarb": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.FastICA(whiten="arbitrary-variance"),
        # ),
        # "Robust&Weight20-80+ICAunit": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.FastICA(whiten="unit-variance"),
        # ),
        # "Robust&Weight20-80+FA": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.FactorAnalysis(max_iter=200),
        # ),
        # "Robust&Weight20-80+dictlars": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(20, 80),
        #     ),
        #     apply_weight,
        #     dec.DictionaryLearning(fit_algorithm="lars", alpha=0.1, n_jobs=-1),
        # ),
        "Robust&Weight20-80+dictcd": (
            RobustScaler(
                with_centering=True,
                with_scaling=True,
                quantile_range=(20, 80),
            ),
            apply_weight,
            dec.DictionaryLearning(fit_algorithm="cd", alpha=0.1, n_jobs=-1),
        ),
        # "Robust&Weight10-90": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(10, 90),
        #     ),
        #     apply_weight,
        # ),
        # "Robust&Weight05-95": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(5, 95),
        #     ),
        #     apply_weight,
        # ),
        # "Robust&Weight01-99": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(1, 99),
        #     ),
        #     apply_weight,
        # ),
        # "Robust&Weight00-100": (
        #     RobustScaler(
        #         with_centering=True,
        #         with_scaling=True,
        #         quantile_range=(0, 100),
        #     ),
        #     apply_weight,
        # ),
    }

    for pre, pp in preprocs.items():
        print("-" * 30)
        print(pre)
        df = selected.copy()
        for p in pp:
            print(f"    - {p!r}")
            df = p.fit_transform(df)
        hop = hopkins(df, sampling_size=150)
        print(f"{pre}, Hopkins: {hop:.5f}")

    seltrans = pd.DataFrame(df, columns=selected.columns, index=selected.index)

    # %%
    ## DEFINE MODELS AD ALGORITHMS

    # define the cluster algorithm to use/test
    models = {
        # "KMEANS": {
        #     "cls": KMeansInterp,
        #     "opts": [
        #         dict(
        #             n_clusters=k,
        #             n_init=20,
        #             max_iter=600,
        #             random_state=23,
        #             ordered_feature_names=selected.columns,
        #             feature_importance_method="wcss_min",
        #         )
        #         for k in range(2, 25)
        #     ],
        #     "pre-processing": [
        #         # PowerTransformer(), # using it_ hopkins is reduced to 0.6884 from 0.9273
        #         RobustScaler(
        #             with_centering=True,
        #             with_scaling=True,
        #             quantile_range=(20, 80),
        #         ),
        #         apply_weight,
        #         dec.DictionaryLearning(fit_algorithm="cd", alpha=0.1, n_jobs=-1),
        #     ],
        #     "label": lambda x: (f"kmeans__k-{x['n_clusters']:02d}"),
        #     "title": lambda x: (f"KMeans: k: {x['n_clusters']:2d} "),
        # },
        # "DBSCAN": {
        #     "cls": DBSCAN,
        #     "opts": [
        #         {"eps": eps, "min_samples": ms}
        #         for eps, ms in product(
        #             (0.1, 0.25, 0.5, 0.75, 0.9), (1, 3, 5, 10, 15, 20)
        #         )
        #     ],
        #     "pre-processing": [
        #         RobustScaler(
        #             with_centering=True,
        #             with_scaling=True,
        #             quantile_range=quantile_range,
        #         ),
        #         apply_weight,
        #         dec.DictionaryLearning(fit_algorithm="cd", alpha=0.1, n_jobs=-1),
        #     ],
        #     "label": lambda x: (
        #         f"dbscan__eps-{x['eps']:3.2f}_"
        #         f"ms-{x['min_samples'] if x['min_samples'] else 0:02d}"
        #     ),
        #     "title": lambda x: (
        #         f"DBSCAN: eps: {x['eps']:3.1f} " f"min_samples: {x['min_samples']}"
        #     ),
        # },
        "HDBSCAN": {
            "cls": HDBSCAN,
            "opts": [
                {
                    "min_cluster_size": mcs,
                    "min_samples": ms,
                    "metric": mtr,
                    "p": 0.01 if mtr == "minkowski" else None,
                    "cluster_selection_method": csm,
                }
                for mcs, ms, mtr, csm in product(
                    [2, 3, 5, 7, 10, 12, 15],  # min_cluster_size
                    [None, 1, 2, 3, 5],  # min_samples
                    [
                        "euclidean",
                        # "haversine", # only 2D
                        "cityblock",
                        # "cosine",
                        "l1",
                        "l2",
                        "manhattan",
                        "braycurtis",
                        "canberra",
                        "chebyshev",
                        "correlation",
                        # "dice",   # only boolean vectors
                        # "hamming",   # only boolean vectors
                        # "jaccard",   # only boolean vectors
                        # "kulsinski",   # only boolean vectors
                        # "mahalanobis", # Must provide either V or VI for Mahalanobis distance
                        "minkowski",
                        # "rogerstanimoto",  # only boolean vectors
                        # "russellrao",  # only boolean vectors
                        # "seuclidean",
                        # "sokalmichener",  # only boolean vectors
                        # "sokalsneath",  # only boolean vectors
                        "sqeuclidean",
                        # "yule", # only boolean vectors
                    ],  # metric
                    ["eom", "leaf"],  # cluster_selection_method
                )
            ],
            # "pre-processing": [
            #     RobustScaler(
            #         with_centering=True,
            #         with_scaling=True,
            #         quantile_range=quantile_range,
            #     ),
            #     apply_weight,
            #     dec.DictionaryLearning(fit_algorithm="cd", alpha=0.1, n_jobs=-1),
            # ],
            "label": lambda x: (
                f"hdbscan__mcs-{x['min_cluster_size']:02d}_"
                f"ms-{x['min_samples'] if x['min_samples'] else 0:02d}_"
                f"m-{x['metric']}_"
                f"csm-{x['cluster_selection_method']}"
            ),
            "title": lambda x: (
                f"Scaled/Weighted HDBSCAN: min_cluster_size: {x['min_cluster_size']} "
                f"min_samples: {x['min_samples']}"
                f"metric: {x['metric']}"
                f"csm: {x['cluster_selection_method']}"
            ),
        },
        # "BIRCH": {
        #     "cls": Birch,
        #     "opts": [
        #         dict(
        #             threshold=thr,
        #             branching_factor=bf,
        #             n_clusters=None,
        #         )
        #         for thr, bf, k in product(
        #             (0.1, 0.25, 0.5, 0.75, 0.9, 1.15, 1.5),
        #             (25, 50, 100),
        #             (None, 2, 3, 5, 6, 7, 8, 9, 10, 11, 12, 15, 20, 25),
        #         )
        #     ],
        #     "pre-processing": [
        #         RobustScaler(
        #             with_centering=True,
        #             with_scaling=True,
        #             quantile_range=quantile_range,
        #         ),
        #         apply_weight,
        #         dec.DictionaryLearning(fit_algorithm="cd", alpha=0.1, n_jobs=-1),
        #     ],
        #     "label": lambda x: (
        #         f"birch__t-{x['threshold']:4.2f}_"
        #         f"b-{x['branching_factor']}_"
        #         f"k-{x['n_clusters']}"
        #     ),
        #     "title": lambda x: (
        #         f"Birch t: {x['threshold']:4.2f} "
        #         f"b: {x['branching_factor']} "
        #         f"k: {x['n_clusters']}"
        #     ),
        # },
    }

    # %%
    # Explore clustering algorithms
    cldf = raw.loc[:, sez_cols].copy()
    rdir = paths.rpdir / "all_dictlearning_cd"
    os.makedirs(rdir, exist_ok=True)
    scores, fimps = report(seltrans, "all", models, cldf, rdir)
    scores.sort_values(
        by=["n_clusters", "% covered by clusters"],
        ascending=[True, False],
        inplace=True,
    )
    scores.to_excel(rdir / "scores.xlsx")
    fimps.T.to_excel(rdir / "feature_importances.xlsx")

    min_perc = 50  # save to gpkg only if perc > x
    interesting_scores = scores.where(scores["% covered by clusters"] > min_perc)[
        "long_label"
    ].dropna()
    cldf.loc[:, sez_cols + list(interesting_scores)].to_file(
        rdir / "all_clusters.gpkg", driver="GPKG"
    )

    # %%
    print("Best based on Silhouette (higher is better)")
    print(
        scores.sort_values(by="silhouette", ascending=False).loc[
            :10, ["long_label", "% covered by clusters", "n_clusters"]
        ]
    )

    print("Best based on Calinski Harabasz (higher is better)")
    print(
        scores.sort_values(by="calinski harabasz", ascending=False).loc[
            :10, ["long_label", "% covered by clusters", "n_clusters"]
        ]
    )

    print("Best based on Davies Bouldin (lower is better)")
    print(
        scores.sort_values(by="davies bouldin", ascending=True).loc[
            :10, ["long_label", "% covered by clusters", "n_clusters"]
        ]
    )

    # pietro index
    sc = pd.DataFrame(
        RobustScaler(quantile_range=(0.0, 100.0)).fit_transform(
            scores.loc[:, ["silhouette", "davies bouldin", "calinski harabasz"]]
        ),
        columns=["silhouette", "davies bouldin", "calinski harabasz"],
        index=scores.index,
    )
    print(sc.describe())
    sil, dav, cal = sc.values.T
    #        1   + 1   - (-1)
    pindex = pd.Series(sil + cal - dav, index=scores.index, name="pindex")
    pi = pd.Series(
        RobustScaler(quantile_range=(10, 90))
        .fit_transform(pindex.values.reshape(-1, 1))
        .T[0],
        index=scores.index,
        name="pindex",
    )
    print("Best based on Pietro's index (higher is better)")
    print(
        scores.loc[
            pi.sort_values(ascending=False)[:10].index,
            ["long_label", "% covered by clusters", "n_clusters"],
        ]
    )

    """
    sil = silhouette_score(data, labels)
        dvb = davies_bouldin_score(data, labels)
        cal = calinski_harabasz_score(data, labels)
        return k, inertia, scaledi, sil, dvb, cal
    # Instantiate the clustering model and visualizer
    model = KMeans()
    visualizer = KElbowVisualizer(
        model, k=(4, 12), metric="calinski_harabasz", timings=False
    )

    visualizer.fit(X)  # Fit the data to the visualizer
    visualizer.show()  # Finalize and render the figure

    ## other method
    kms = KMeansInterp(
        n_clusters=5,
        ordered_feature_names=X.columns.tolist(),
        feature_importance_method="wcss_min",  # or 'unsup2sup'
    ).fit(X.values)

    # A dictionary where the key [0] is the cluster label, and [:10] will refer to the first 10 most important features
    kms.feature_importances_[0][:10]
    """

# %%
